import redis_lock
from redis_om import Migrator, get_redis_connection


class LockContextManager:
    """Lock context manager."""

    def __init__(self, lock):
        self.lock = lock

    def __enter__(self):
        try:
            self.lock.acquire(timeout=1)
        except redis_lock.NotAcquired:
            return None
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()


class TranslateClientModule:
    __instance = None

    def __new__(cls, *args, **kwargs):
        if not cls.__instance:
            cls.__instance = super().__new__(cls, *args, **kwargs)
        return cls.__instance

    def __init__(self):
        self.conn = get_redis_connection()
        self.lock = redis_lock.Lock(self.conn, "melon_translate_process_lock")
        Migrator().run()

    def lock_context(self) -> LockContextManager:
        """Return lock context manager."""
        return LockContextManager(self.lock)


translate_module = None



import json
import secrets
import time
from pathlib import Path
from typing import Dict, List, Optional

from melon_translate_client import module
from melon_translate_client.request import fetch_all, fetch_all_par
from melon_translate_client.settings import (
    TRANSLATE_CACHE_VIEW_TTL,
    TRANSLATE_ENABLED,
    TRANSLATE_REFRESH_ENABLED,
    TRANSLATE_TTL_JITTER,
    log,
)
from redis_om import Field, HashModel, NotFoundError, get_redis_connection


def check_view_refresh(cls_method):
    """Check if view is expired and refresh it if needed."""

    def _wrapper(cls, language, key):
        """Wrapper for cls_method."""
        fetched_key = cls_method(cls, language, key)

        if TRANSLATE_REFRESH_ENABLED and isinstance(fetched_key, TranslateKey):
            views = []

            for vw in fetched_key.views.split(","):
                query = (TranslateView.name == vw) & (
                    TranslateView.language == language
                )
                try:
                    view = TranslateView.find(query).first()
                    if view and view.expired():
                        log.info(f"FOUND VIEW EXPIRED - REFRESHING {view.name}")
                        views.append(view.name)
                except NotFoundError:
                    log.info(f"View `{vw}` not found for language `{language}`")
                    TranslateView(name=vw, language=language).refresh()

            if views:
                log.info(f"Refreshing views `{views}` for language `{language}`")
                TranslateView.refresh_bulk(language, views)

        return fetched_key.translation

    return _wrapper


class TranslateKey(HashModel):
    """Translate model."""

    snake_name: Optional[str] = Field(index=True)
    id_name: Optional[str] = Field(index=True)
    views: Optional[str] = Field(index=True, full_text_search=True)
    occurrences: Optional[str] = Field(index=True, full_text_search=True)
    usage_context: Optional[str]
    flags: Optional[str] = ""
    translation: Optional[str]
    language: Optional[str] = Field(index=True)
    updated_at: int = Field(default_factory=lambda: int(time.time()))

    class Meta:
        global_key_prefix = "melon_translate"
        model_key_prefix = "translate_key"

    @classmethod
    @check_view_refresh
    def by_key(cls, language: str, key: str) -> "TranslateKey":
        """Finds translation by snake_name or id_name. In case of not found, returns key itself."""
        qs = cls.find(
            ((cls.snake_name == key) | (cls.id_name == key))
            & (cls.language == language)
        )

        for item in qs:
            if item.snake_name == key or item.id_name == key:
                # NOTE: Actual result received from the client invocation will be translation itself due to the decorator.
                return item
        else:
            raise NotFoundError

    @classmethod
    def from_svc(cls, key_obj: dict, pipeline) -> Optional["TranslateKey"]:
        _id = key_obj.get("id")
        if not _id:
            return None

        _id = str(_id)
        _key = key_obj.get("key")
        _data = cls(
            pk=_id,
            snake_name=_key.get("snake_name"),
            id_name=_key.get("id_name"),
            views=",".join(_key.get("views")),
            translation=key_obj.get("translation"),
            language=key_obj.get("language"),
        )
        _data.save(pipeline=pipeline)
        return _data


class TranslateView(HashModel):
    """View model."""

    name: str = Field(index=True)
    language: str = Field(index=True)
    updated_at: int = Field(default_factory=lambda: int(time.time()))

    class Meta:
        global_key_prefix = "melon_translate"
        model_key_prefix = "translate_view"

    @classmethod
    def refresh_bulk(cls, language: str, view_names: List[str], page_size=1000) -> int:
        """Refreshes all views in bulk."""
        n_writes = 0
        pipeline = get_redis_connection().pipeline()

        with module.translate_module.lock_context() as ctx:
            if not ctx:
                log.error(
                    f"Failed to acquire lock for refresh_bulk `{language}`:`{view_names}`"
                )
                return 0

            for view in view_names:
                try:
                    view = TranslateView.find(
                        (TranslateView.name == view)
                        & (TranslateView.language == language)
                    ).first()
                    view.updated_at = int(time.time())
                    view.save(pipeline=pipeline)
                except NotFoundError:
                    TranslateView(
                        name=view,
                        language=language,
                        updated_at=int(time.time()),
                    ).save(pipeline=pipeline)
            pipeline.execute()

            for result_list in fetch_all_par(
                [language], view_names, page_size=page_size
            ):
                keys = [TranslateKey.from_svc(obj, pipeline) for obj in result_list]
                n_writes += len(keys)
                pipeline.execute()

        return n_writes

    def refresh(self, page_size=1000) -> int:
        """Fetches all keys for current view and language and caches them."""
        pipeline = get_redis_connection().pipeline()

        with module.translate_module.lock_context() as ctx:
            if not ctx:
                log.error(f"Failed to acquire lock for view `{self.name}` refresh")
                return 0

            fetched_keys = fetch_all(
                language=self.language,
                views=[self.name],
                page_size=page_size,
            )

        for obj in fetched_keys:
            TranslateKey.from_svc(obj, pipeline=pipeline)

        self.updated_at = int(time.time())
        self.save(pipeline=pipeline)
        pipeline.execute()

        return len(fetched_keys)

    def expired(self) -> bool:
        """Check if view is expired."""
        jitter = secrets.randbelow(TRANSLATE_TTL_JITTER + 1)
        expected_time = int(time.time()) - (TRANSLATE_CACHE_VIEW_TTL + jitter)
        return self.updated_at < expected_time

    def keys(self, default_attr="snake_name") -> Dict[str, str]:
        """Get all keys corresponding to a view."""
        all_keys = TranslateKey.find(
            (TranslateKey.views % self.name) & (TranslateKey.language == self.language)
        ).all()

        return {getattr(key, default_attr): key.translation for key in all_keys}


if TRANSLATE_ENABLED:
    setattr(module, "translate_module", module.TranslateClientModule())



import concurrent.futures
import time
from itertools import product
from math import ceil
from typing import List, Optional
from urllib.parse import urlparse

import gevent
import requests
from melon_translate_client.settings import (
    TRANSLATE_REQUESTS_TIMEOUT,
    TRANSLATE_URL,
    TRANSLATE_WORKER_POOL_SIZE,
    log,
)


def translate_url(language, target_url: str = None) -> str:
    """Get target translate service URL."""
    target_url = target_url or TRANSLATE_URL

    _url = urlparse(f"{target_url}/api/v2/translations/{language}/")
    return _url.geturl()


def build_params(
    views: List[str],
    timestamp: int = None,
    page_size: int = None,
    page: int = None,
) -> dict:
    """Build HTTP params."""
    params = {}
    if timestamp:
        params["timestamp"] = timestamp

    if views:
        params["views"] = views

    if page_size and page:
        params["page_size"] = page_size
        params["page"] = page

    return params


def fetch(url):
    """Fetch data from url."""
    response = requests.get(url, timeout=TRANSLATE_REQUESTS_TIMEOUT)
    return url, response.json()


def fetch_all(
    language: str, views: List[str], page_size: int = 500, timestamp=None
) -> Optional[list]:
    """Return all pages for specified query."""
    url = translate_url(language)

    params = build_params(views, timestamp, page_size, 1)
    headers = {"Cache-Control": "no-cache"}

    response = requests.get(
        url,
        params=params,
        timeout=TRANSLATE_REQUESTS_TIMEOUT,
        headers=headers,
    )
    full_url = response.url
    response = response.json()

    if not response.get("results"):
        log.error(f"[ERROR] url={url}, params={params}, response={response}")
        return None

    results = response.get("results")
    count = response.get("count")

    pages = ceil(count / page_size)
    urls = [full_url.replace("page=1", f"page={idx}") for idx in range(2, pages + 1)]

    greenlets = [gevent.spawn(fetch, url) for url in urls]
    for greenlet in gevent.iwait(greenlets):
        url, response_json = greenlet.value
        log.debug(
            f"answer from {url} received {len(response_json.get('results', []))} objects\n",
            url,
        )
        results += response_json.get("results", [])

    # NOTE: Should be optimized.
    for response in results:
        response["language"] = language

    return results


def fetch_all_par(
    languages: List[str],
    # NOTE: views_par is a list of strings chained together with OR operator in parallel requests mode
    views_par: List[str],
    page_size: int = 1000,
):
    """Return all pages for for product of languages and views."""
    pairs = list(product(languages, views_par))
    timestamp = int(time.time())

    # TODO: Use gevent here as well, no need for context switching.
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=TRANSLATE_WORKER_POOL_SIZE or None
    ) as executor:
        result_future = (
            executor.submit(
                lambda p: fetch_all(p[0], [p[1]], page_size, timestamp), pair
            )
            for pair in pairs
        )

        for future in concurrent.futures.as_completed(result_future):
            try:
                result = future.result()
            except Exception as exc:  # TODO: find correct exception to catch
                log.exception(exc)
            else:
                if result:
                    yield result



import logging

from decouple import config

# NOTE: Logging configuration
log = logging.getLogger("melon_translate_client")

# NOTE: Mandatory settings
TRANSLATE_URL = config(
    "TRANSLATE_URL", default="https://translate.dev.getmelon.dev"
)

# NOTE: Optional settings
TRANSLATE_ENABLED = config("TRANSLATE_ENABLED", default=False, cast=bool)
TRANSLATE_REFRESH_ENABLED = config("TRANSLATE_REFRESH_ENABLED", default=False, cast=bool)
TRANSLATE_REQUESTS_TIMEOUT = config("TRANSLATE_REQUESTS_TIMEOUT", default=30, cast=int)
TRANSLATE_CACHE_VIEW_TTL = config("TRANSLATE_CACHE_VIEW_TTL", default=-1, cast=int)
TRANSLATE_TTL_JITTER = config("TRANSLATE_TTL_JITTER", default=0, cast=int)

# NOTE: This is the number of worker threads that will be spawned to process translation requests.
# To calculate exact number of threads that will be spawned when size is `0` on Kubernetes we have
# to look at the K8S limits for the container in which we spawn the worker. If there is no limit, the
# container will see the host node CPU count and use following formula to calculate the number of threads:
#  `cpu_count + 4` - for both IO and CPU bound tasks which release GIL.
TRANSLATE_WORKER_POOL_SIZE = config("TRANSLATE_WORKER_POOL_SIZE", default=0, cast=int)
